from enum import IntEnum

__all__ = ["Label"]


class Label(IntEnum):
    """
    Anomalous classification labels.

    Three types of labels:

      * -1 for anomalies, referenced either as `Label.ANOMALY` or as `Label.A`,
      * 0 for unknowns: `Label.UNKNOWN` or `Label.U`,
      * 1 for regular data: `Label.REGULAR` or `Label.R`.
    """

    ANOMALY = -1
    A = ANOMALY
    UNKNOWN = 0
    U = UNKNOWN
    REGULAR = 1
    R = REGULAR



import numpy as np

__all__ = ["average_path_length"]


def _average_path_length(n):
    """
    Average path length formula.
    """
    # Thank you Matwey.
    return 2.0 * (np.log(n) + np.euler_gamma + 1 / (2 * n) - 1 / (12 * n**2) - 1.0)


def average_path_length(n):
    """
    Average path length computation.

    Parameters
    ----------
    n
        Either array of tree depths to computer average path length of
        or one tree depth scalar.

    Returns
    -------
    Average path length.
    """
    if np.isscalar(n):
        if n <= 1:
            apl = 0.0
        else:
            apl = _average_path_length(n)
    else:
        n = np.asarray(n)
        apl = np.zeros_like(n)
        apl[n > 1] = _average_path_length(n[n > 1])

    return apl



import numpy as np

from .evaluator import ForestEvaluator
from .utils import average_path_length

# Not very useful classes at the moment.
# Implemented purely with the educational purpose.


class RandomLimeForest:
    def __init__(self, trees=100, subsamples=256, depth=None, seed=0):
        self.subsamples = subsamples
        self.trees = trees
        self.depth = depth

        self.seedseq = np.random.SeedSequence(seed)
        self.rng = np.random.default_rng(seed)

        self.estimators = []
        self.n = 0

    def fit(self, data):
        n = data.shape[0]
        self.n = n
        self.subsamples = self.subsamples if n > self.subsamples else n

        self.depth = self.depth or int(np.ceil(np.log2(self.subsamples)))

        self.estimators = [None] * self.trees
        seeds = self.seedseq.spawn(self.trees)
        for i in range(self.trees):
            subs = self.rng.choice(n, self.subsamples)
            gen = RandomLimeGenerator(data[subs, :], self.depth, seeds[i])
            self.estimators[i] = gen.pine

        return self

    def mean_paths(self, data):
        means = np.zeros(data.shape[0])
        for ti in range(self.trees):
            path = self.estimators[ti].paths(data)
            means += path

        means /= self.trees
        return means

    def scores(self, data):
        means = self.mean_paths(data)
        return -(2 ** (-means / average_path_length(self.subsamples)))


class RandomLime:
    def __init__(self, features, selectors, values):
        self.features = features
        self.len = selectors.shape[0]

        # Two complementary arrays.
        # Selectors select feature to branch on.
        self.selectors = selectors
        # Values either set the deciding feature value or set the closing path length
        self.values = values

    def _get_one_path(self, key):
        i = 1
        while 2 * i < self.selectors.shape[0]:
            f = self.selectors[i]
            if f < 0:
                break

            if key[f] <= self.values[i]:
                i = 2 * i
            else:
                i = 2 * i + 1

        return self.values[i]

    def paths(self, x):
        n = x.shape[0]
        paths = np.empty(n)
        for i in range(n):
            paths[i] = self._get_one_path(x[i, :])

        return paths


class RandomLimeGenerator:
    def __init__(self, sample, depth, seed=0):
        self.depth = depth
        self.features = sample.shape[1]
        self.length = 1 << (depth + 1)
        self.rng = np.random.default_rng(seed)
        self.selectors = np.full(self.length, -1, dtype=np.int32)
        self.values = np.full(self.length, 0, dtype=np.float64)

        self._populate(1, sample)

        self.pine = RandomLime(self.features, self.selectors, self.values)

    def _populate(self, i, sample):
        if sample.shape[0] == 1:
            self.values[i] = np.floor(np.log2(i))
            return

        if self.length <= 2 * i:
            self.values[i] = np.floor(np.log2(i)) + average_path_length(sample.shape[0])

            return

        selector = self.rng.integers(self.features)
        self.selectors[i] = selector

        minval = np.min(sample[:, selector])
        maxval = np.max(sample[:, selector])
        if minval == maxval:
            self.selectors[i] = -1
            self.values[i] = np.floor(np.log2(i)) + average_path_length(sample.shape[0])

            return

        value = self.rng.uniform(minval, maxval)
        self.values[i] = value

        self._populate(2 * i, sample[sample[:, selector] <= value])
        self._populate(2 * i + 1, sample[sample[:, selector] > value])


class LimeEvaluator(ForestEvaluator):
    def __init__(self, pine_forest):
        pines = pine_forest.estimators
        self.trees = len(pines)
        if self.trees < 1:
            raise ValueError("a forest without trees?")

        selectors, node_offsets, leaf_offsets = self.combine_selectors([self.extract_selectors(pine) for pine in pines])

        super(LimeEvaluator, self).__init__(
            samples=pine_forest.subsamples, selectors=selectors, node_offsets=node_offsets, leaf_offsets=leaf_offsets
        )

    @classmethod
    def extract_selectors(cls, pine):
        selectors = np.zeros((pine.len,), dtype=cls.selector_dtype)

        mapping = np.full((pine.len,), -1, dtype=np.int32)
        current = 0
        for i in range(1, pine.len):
            if pine.selectors[i] != -1 or pine.values[i] != 0:
                mapping[i] = current
                current += 1

        selectors = selectors[:current]

        for i in range(1, pine.len):
            if pine.selectors[i] != -1 or pine.values[i] != 0:
                current = mapping[i]

                feature = pine.selectors[i]
                selectors[current]["feature"] = feature
                selectors[current]["value"] = pine.values[i]

                if 2 * i >= pine.len:
                    continue

                selectors[current]["left"] = mapping[2 * i]
                selectors[current]["right"] = mapping[2 * i + 1]

        return selectors



import numpy as np
from sklearn.tree._tree import DTYPE as TreeDTYPE  # noqa

from .coniferest import Coniferest, ConiferestEvaluator
from .label import Label
from .utils import average_path_length

__all__ = ["PineForest"]


class PineForest(Coniferest):
    """
    Pine Forest for active anomaly detection.

    Pine Forests are filtering isolation forests. That's a simple concept
    of incorporating prior knowledge about what is anomalous and what is not.

    Standard fit procedure with two parameters works exactly the same as the
    isolation forests' one. It differs when we supply additional parameter
    `labels`, then the behaviour changes. At that case fit generates additional
    not only `n_trees` but with additional `n_spare_trees` and then filters out
    `n_spare_trees`, leaving only those `n_trees` that deliver better scores
    for the data known to be anomalous.

    Parameters
    ----------
    n_trees : int, optional
        Number of trees to keep for estimating anomaly scores.

    n_subsamples : int, optional
        How many subsamples should be used to build every tree.

    max_depth : int or None, optional
        Maximum depth of every tree. If None, `log2(n_subsamples)` is used.

    n_spare_trees : int, optional
        Number of trees to generate additionally for further filtering.

    regenerate_trees : bool, optional
        Should we throughout all the trees during retraining or should we
        mix old trees with the fresh ones. False by default, so we mix.

    weight_ratio : float, optional
        What is the relative weight of false positives relative to true
        positives (i.e. we are not interested in negatives in anomaly
        detection, right?). The weight is used during the filtering
        process.

    n_jobs : int, optional
        Number of threads to use for scoring. If None - number of CPUs is used.

    random_seed : int or None, optional
        Random seed. If None - random seed is used.
    """

    def __init__(
        self,
        n_trees=100,
        n_subsamples=256,
        max_depth=None,
        n_spare_trees=400,
        regenerate_trees=False,
        weight_ratio=1.0,
        n_jobs=None,
        random_seed=None,
        sampletrees_per_batch=1 << 20,
    ):
        super().__init__(
            trees=[],
            n_subsamples=n_subsamples,
            max_depth=max_depth,
            n_jobs=n_jobs,
            random_seed=random_seed,
            sampletrees_per_batch=sampletrees_per_batch,
        )
        self.n_trees = n_trees
        self.n_spare_trees = n_spare_trees
        self.weight_ratio = weight_ratio
        self.regenerate_trees = regenerate_trees

        self.evaluator = None

    def _expand_trees(self, data, n_trees):
        """
        Expand the forest. Grow new trees.

        Parameters
        ----------
        data
            Data to build the trees from.

        n_trees
            What tree population are we aiming at?

        Returns
        -------
        None
        """
        n = n_trees - len(self.trees)
        if n > 0:
            self.trees.extend(self.build_trees(data, n))

    def _contract_trees(self, known_data, known_labels, n_trees):
        """
        Contract the forest with a chainsaw. According to the known data.

        Parameters
        ----------
        known_data
            Array with features of objects we know labels about.

        known_labels
            Array with labels. Of Label type.

        n_trees
            Aiming tree population after contraction.

        Returns
        -------
        None
        """
        n_filter = len(self.trees) - n_trees
        if n_filter > 0:
            self.trees = self.filter_trees(
                trees=self.trees,
                data=known_data,
                labels=known_labels,
                n_filter=n_filter,
                weight_ratio=self.weight_ratio,
            )

    def fit(self, data, labels=None):
        """
        Build the trees with the data `data`.

        Parameters
        ----------
        data
            Array with feature values of objects.

        labels
            Optional. Labels of objects. May be regular, anomalous or unknown.
            See `Label` data for details.

        Returns
        -------
        self
        """
        # If no labels were supplied, train with them.
        if labels is None:
            self.fit_known(data)
            return self

        # Otherwise select known data, and train on it.
        labels = np.asarray(labels)
        index = labels != Label.UNKNOWN
        return self.fit_known(data, data[index, :], labels[index])

    def fit_known(self, data, known_data=None, known_labels=None):
        """
        The same `fit` but with a bit of different API. Known data and labels
        are separated from training data for time and space optimality. High
        chances are that `known_data` is much smaller that `data`. At that case
        it is not reasonable to hold the labels for whole `data`.

        Parameters
        ----------
        data
            Training data (array with feature values) to build trees with.

        known_data
            Feature values of known data.

        known_labels
            Labels of known data.

        Returns
        -------
        self
        """

        known_data, known_labels = self._validate_known_data(known_data, known_labels)

        if self.regenerate_trees:
            self.trees = []

        if (
            known_data is None
            or len(known_data) == 0
            or known_labels is None
            or len(known_labels) == 0
            or np.all(known_labels == Label.UNKNOWN)
        ):
            self._expand_trees(data, self.n_trees)
        else:
            self._expand_trees(data, self.n_trees + self.n_spare_trees)
            self._contract_trees(known_data, known_labels, self.n_trees)

        self.evaluator = ConiferestEvaluator(self)
        return self

    # Made non-static to make sure we always use it in a way that makes inheritance with truly non-static methods
    # possible.
    def filter_trees(self, trees, data, labels, n_filter, weight_ratio=1):
        """
        Filter the trees out.

        Parameters
        ----------
        trees
            Trees to filter.

        n_filter
            Number of trees to filter out.

        data
            The labeled objects themselves.

        labels
            The labels of the objects. -1 is anomaly, 1 is not anomaly, 0 is uninformative.

        weight_ratio
            Weight of the false positive experience relative to false negative. Defaults to 1.
        """
        data = np.asarray(data, dtype=TreeDTYPE)

        n_samples, _ = data.shape
        n_trees = len(trees)

        heights = np.empty(shape=(n_samples, n_trees))
        for tree_index in range(n_trees):
            tree = trees[tree_index]
            leaves_index = tree.apply(data)
            n_samples_leaf = tree.n_node_samples[leaves_index]
            n_samples_leaf = n_samples_leaf.astype(dtype=np.float64)

            heights[:, tree_index] = (
                np.ravel(tree.decision_path(data).sum(axis=1)) + average_path_length(n_samples_leaf) - 1
            )

        weights = labels.copy()
        weights[labels == Label.REGULAR] = weight_ratio * Label.REGULAR
        weighted_paths = (heights * np.reshape(weights, (-1, 1))).sum(axis=0)
        indices = weighted_paths.argsort()[n_filter:]

        return [trees[i] for i in indices]

    def score_samples(self, samples):
        """
        Computer scores for the supplied data.

        Parameters
        ----------
        samples
            Feature values to compute scores on.

        Returns
        -------
        Array with computed scores.
        """
        return self.evaluator.score_samples(samples)

    def feature_signature(self, x):
        return self.evaluator.feature_signature(x)

    def feature_importance(self, x):
        return self.evaluator.feature_importance(x)

    def apply(self, x):
        """
        Apply the forest to X, return leaf indices.

        Parameters
        ----------
        x : ndarray shape (n_samples, n_features)
            2-d array with features.

        Returns
        -------
        x_leafs : ndarray of shape (n_samples, n_estimators)
            For each datapoint x in X and for each tree in the forest,
            return the index of the leaf x ends up in.
        """
        return self.evaluator.apply(x)



from .coniferest import Coniferest, ConiferestEvaluator

__all__ = ["IsolationForest"]


class IsolationForest(Coniferest):
    """
    Isolation forest.

    This is a reimplementation of sklearn.ensemble.IsolationForest,
    which trains and evaluates much faster. It also supports multi-threading
    for evaluation (sample scoring).

    Parameters
    ----------
    n_trees : int, optional
        Number of trees in forest to build.

    n_subsamples : int, optional
        Number of subsamples to use for building the trees.

    max_depth : int or None, optional
        Maximal tree depth. If None, `log2(n_subsamples)` is used.

    n_jobs : int or None, optional
        Number of threads to use for evaluation. If None, use all available CPUs.

    random_seed : int or None, optional
        Seed for reproducibility. If None, random seed is used.
    """

    def __init__(
        self,
        n_trees=100,
        n_subsamples=256,
        max_depth=None,
        n_jobs=None,
        random_seed=None,
        sampletrees_per_batch=1 << 20,
    ):
        super().__init__(
            trees=[],
            n_subsamples=n_subsamples,
            max_depth=max_depth,
            n_jobs=n_jobs,
            random_seed=random_seed,
            sampletrees_per_batch=sampletrees_per_batch,
        )
        self.n_trees = n_trees
        self.evaluator = None

    def fit(self, data, labels=None):
        """
        Build the trees based on data.

        Parameters
        ----------
        data
            2-d array with features.

        labels
            Unused. Defaults to None.

        Returns
        -------
        self
        """
        self.trees = self.build_trees(data, self.n_trees)
        self.evaluator = ConiferestEvaluator(self)
        return self

    def score_samples(self, samples):
        """
        Compute scores for given samples.

        Parameters
        ----------
        samples
            2-d array with features.

        Returns
        -------
        1-d array with scores.
        """
        return self.evaluator.score_samples(samples)

    def fit_known(self, data, known_data=None, known_labels=None):
        return self.fit(data)

    def feature_signature(self, x):
        return self.evaluator.feature_signature(x)

    def feature_importance(self, x):
        return self.evaluator.feature_importance(x)

    def apply(self, x):
        """
        Apply the forest to X, return leaf indices.

        Parameters
        ----------
        x : ndarray shape (n_samples, n_features)
            2-d array with features.

        Returns
        -------
        x_leafs : ndarray of shape (n_samples, n_estimators)
            For each datapoint x in X and for each tree in the forest,
            return the index of the leaf x ends up in.
        """
        return self.evaluator.apply(x)



from __future__ import annotations  # "|" syntax is not available in Python 3.9

IMPORT_ERROR_MESSAGE = (
    "`datasets' library is required to load PLAsTiCC dataset, install it with `pip install datasets` "
    "or reinstall the package with `pip install coniferest[datasets]`"
)


def plasticc_gp(**kwargs):
    """
    Load PLAsTiCC dataset of GP-approximated light curves.

    It is a subset of the PLAsTiCC dataset, light curves are approximated
    with vector Gaussian processes (Kornilov et al., 2023). Features
    are these GP approximations and GP parameters, all normalized.

    The dataset is hosted on Hugging Face, `datasets' package is required
    to load it.

    Adopted from Ishida et al. 2021:
    https://ui.adsabs.harvard.edu/abs/2021A%26A...650A.195I/abstract

    Parameters
    ----------
    **kwargs
        Arbitrary keyword arguments passed to `datasets.load_dataset`.

    Returns
    -------
    data : 2-D numpy.ndarray of float32
        2-D array of features.
    metadata : 1-D numpy.ndarray of bool
        1-D array of anomaly labels, True for anomalies.
    """

    try:
        from datasets import load_dataset
    except ImportError:
        raise ImportError(IMPORT_ERROR_MESSAGE)

    dataset = load_dataset("snad-space/plasticc-gp", **kwargs)
    df = dataset["train"].to_pandas()
    features_columns = [name for name in df.columns if name.startswith("feature")]
    data = df[features_columns].to_numpy()
    # The source file uses 1 for anomaly and 0 for nominals,
    # while we need Label.ANOMALY == -1 and Label.REGULAR == 1.
    metadata = 1 - 2 * df["answer"].to_numpy(dtype=int)
    return data, metadata



from onnxconverter_common.registration import register_converter


def get_default_attribute_pairs():
    attrs = {}
    attrs["post_transform"] = "NONE"
    attrs["n_targets"] = 1
    attrs["nodes_treeids"] = []
    attrs["nodes_nodeids"] = []
    attrs["nodes_featureids"] = []
    attrs["nodes_modes"] = []
    attrs["nodes_values"] = []
    attrs["nodes_truenodeids"] = []
    attrs["nodes_falsenodeids"] = []
    attrs["target_treeids"] = []
    attrs["target_nodeids"] = []
    attrs["target_ids"] = []
    attrs["target_weights"] = []
    return attrs


def get_leaf_weight(tree_id, node_id, parents, model):
    evaluator = model.evaluator
    selector_id = evaluator.node_offsets[tree_id].astype(int) + node_id
    selector = evaluator.selectors[selector_id]
    value = selector["value"]
    # Generally, selector["leaf"] should be -1 for leafs,
    # but for AADForest it stores related weight index.
    leaf_id = selector["left"]
    weight = evaluator.weights[leaf_id]

    return weight * value


def add_node(attr_pairs, tree_id, node_id, feature_id, mode, value, true_child_id, false_child_id):
    attr_pairs["nodes_treeids"].append(tree_id)
    attr_pairs["nodes_nodeids"].append(node_id)
    attr_pairs["nodes_featureids"].append(feature_id)
    attr_pairs["nodes_modes"].append(mode)
    attr_pairs["nodes_values"].append(value)
    attr_pairs["nodes_truenodeids"].append(true_child_id)
    attr_pairs["nodes_falsenodeids"].append(false_child_id)


def add_leaf(attr_pairs, tree_id, node_id, mode, weight):
    add_node(attr_pairs, tree_id, node_id, 0, mode, 0.0, 0, 0)

    attr_pairs["target_treeids"].append(tree_id)
    attr_pairs["target_nodeids"].append(node_id)
    attr_pairs["target_ids"].append(0)
    attr_pairs["target_weights"].append(weight)


def add_tree_to_attribute_pairs(attr_pairs, tree, tree_id, model):
    parents = {}

    for i in range(tree.node_count):
        node_id = i

        if tree.children_left[i] > i or tree.children_right[i] > i:
            mode = "BRANCH_LEQ"
            feat_id = tree.feature[i]
            threshold = tree.threshold[i]
            left_child_id = int(tree.children_left[i])
            right_child_id = int(tree.children_right[i])
            parents[left_child_id] = i
            parents[right_child_id] = i

            add_node(attr_pairs, tree_id, node_id, feat_id, mode, threshold, left_child_id, right_child_id)
        else:
            mode = "LEAF"
            weight = get_leaf_weight(tree_id, node_id, parents, model)

            add_leaf(attr_pairs, tree_id, node_id, mode, weight)


def convert_aadforest(scope, operator, container):
    model = operator.raw_operator

    attr_pairs = get_default_attribute_pairs()

    for i, tree in enumerate(model.trees):
        add_tree_to_attribute_pairs(attr_pairs, tree, i, model)

    container.add_node(
        "TreeEnsembleRegressor",
        operator.input_full_names,
        operator.output_full_names,
        op_domain="ai.onnx.ml",
        name=scope.get_unique_operator_name("TreeEnsembleRegressor"),
        **attr_pairs,
    )


register_converter("AADForest", convert_aadforest)



from onnxconverter_common.data_types import FloatTensorType
from onnxconverter_common.registration import register_shape_calculator
from onnxconverter_common.utils import (
    check_input_and_output_numbers,
    check_input_and_output_types,
)


def calculate_aadforest_output_shapes(operator):
    check_input_and_output_numbers(operator, input_count_range=1, output_count_range=1)
    check_input_and_output_types(operator, good_input_types=[FloatTensorType])
    N = operator.inputs[0].type.shape[0]

    operator.outputs[0].type = FloatTensorType(shape=[N])


register_shape_calculator("AADForest", calculate_aadforest_output_shapes)



import webbrowser

import click

from coniferest.datasets import Label


class _LabelChoice(click.Choice):
    """Label choice class for click

    Accepts (case-insensitive):
        * -1 / a / anomaly / yes
        * 1 / r / regular / no
        * 0 / u / unknown
    """

    def __init__(self):
        super().__init__(Label)

    def normalize_choice(self, choice, ctx):
        del ctx
        if isinstance(choice, Label):
            return choice
        if choice.lower() == "y" or choice.lower() == "yes":
            return Label.ANOMALY
        if choice.lower() == "n" or choice.lower() == "no":
            return Label.REGULAR
        try:
            return int(choice)
        except ValueError:
            pass
        try:
            return getattr(Label, choice.upper())
        except AttributeError:
            pass


def prompt_decision_callback(metadata, data, session) -> Label:
    """
    Prompt user to label the object as anomaly or regular.

    If user sends keyboard interrupt, terminate the session.
    """
    try:
        return click.prompt(
            text=f"Is {metadata} an anomaly? ([A]nomaly / yes, [R]egular / no, [U]nknown)",
            type=_LabelChoice(),
            show_choices=False,
        )
    except click.Abort:
        session.terminate()
        return Label.UNKNOWN


def viewer_decision_callback(metadata, data, session) -> Label:
    """
    Open SNAD Viewer for ZTF DR object. Metadata must be ZTF DR object ID.
    """
    url = "https://ztf.snad.space/view/{}".format(metadata)

    try:
        webbrowser.get().open_new_tab(url)
    except webbrowser.Error:
        click.echo("Check {} for details".format(url))

    return prompt_decision_callback(metadata, data, session)


class TerminateAfter:
    """
    Terminate session after given number of iterations.

    This callback to be used as "on decision callback":
    Session(..., on_decision_callbacks=[TerminateAfter(budget)])

    Parameters
    ----------
    budget : int
        Number of iterations after which session will be terminated.
    """

    def __init__(self, budget: int):
        self.budget = budget
        self.iteration = 0

    def __call__(self, metadata, data, session) -> None:
        self.iteration += 1
        if self.iteration >= self.budget:
            session.terminate()


class TerminateAfterNAnomalies:
    """
    Terminate session after given number of newly labeled anomalies.

    This callback to be used as "on decision callback":
    Session(..., on_decision_callbacks=[TerminateAfter(budget)])

    Parameters
    ----------
    budget : int
        Number of anomalies to stop after.
    """

    def __init__(self, budget: int):
        self.budget = budget
        self.anomalies_count = 0

    def __call__(self, label, _data, session) -> None:
        self.anomalies_count += label == Label.ANOMALY
        if self.anomalies_count >= self.budget:
            session.terminate()



from onnxconverter_common import onnx_ex


def get_maximum_opset_supported():
    from . import __max_supported_opset__

    return min(__max_supported_opset__, onnx_ex.get_maximum_opset_supported())



from typing import Callable, Optional

import numpy as np

from coniferest.coniferest import Coniferest
from coniferest.label import Label
from coniferest.session import Session
from coniferest.session.callback import TerminateAfter, TerminateAfterNAnomalies


class OracleSession(Session):
    """Automated session to run experiments with labeled data.

    Parameters
    ----------
    data : np.ndarray, 2-D
        Array with feature values of objects.
    labels : np.ndarray, 1-D
        Array with true labels, of Label or int type
    model : Coniferest
        Anomaly detection model to use
    max_iterations : int
        Maximum number of asked decisions
    max_anomalies : int
        Maximum number of anomalies to search for
    on_refit_callbacks : list of callable, or callable, or None, optional
        Functions to be called when model is refitted, default is empty list.
    on_decision_callbacks : list of callable, or callable, or None, optional
        Functions to be called when expert decision is made, default is empty list.

    Also see methods and attributes from the base `Session` class
    """

    def __init__(
        self,
        data: np.ndarray,
        labels: np.ndarray,
        *,
        model: Coniferest,
        max_iterations: int,
        max_anomalies: int,
        on_refit_callbacks=None,
        on_decision_callbacks=None,
    ):
        if on_decision_callbacks is None:
            on_decision_callbacks = []
        elif isinstance(on_decision_callbacks, Callable):
            on_decision_callbacks = [
                on_decision_callbacks,
            ]

        super().__init__(
            data=data,
            metadata=labels,
            model=model,
            # Session.metadata is labels, so we just use this candidate metadata is the true label
            decision_callback=lambda label, _features, _self: label,
            on_refit_callbacks=on_refit_callbacks,
            on_decision_callbacks=[
                TerminateAfter(max_iterations),
                TerminateAfterNAnomalies(max_anomalies),
            ]
            + on_decision_callbacks,
        )


def create_oracle_session(
    data: np.ndarray,
    labels: np.ndarray[int],
    *,
    model: Coniferest,
    max_iterations: Optional[int] = None,
    on_refit_callbacks=None,
    on_decision_callbacks=None,
) -> OracleSession:
    """Create an automated session to run experiments with labeled data.

    Parameters
    ----------
    data : np.ndarray, 2-D
        Array with feature values of objects.
    labels : np.ndarray, 1-D
        Array with true labels, of Label or int type
    model : Coniferest
        Anomaly detection model to use
    max_iterations : int or None, optional
        Maximum number of asked decisions. Default is 5 times the number of anomalies.
    on_refit_callbacks : list of callable, or callable, or None, optional
        Functions to be called when model is refitted, default is empty list.
    on_decision_callbacks : list of callable, or callable, or None, optional
        Functions to be called when expert decision is made, default is empty list.

    Returns
    -------
    OracleSession
    """
    n_anomalies = np.sum(labels == Label.ANOMALY)
    max_iterations = min(n_anomalies * 5.0, len(labels)) if max_iterations is None else max_iterations

    return OracleSession(
        data,
        labels,
        model=model,
        max_iterations=max_iterations,
        max_anomalies=n_anomalies,
        on_refit_callbacks=on_refit_callbacks,
        on_decision_callbacks=on_decision_callbacks,
    )



import numpy as np

IMPORT_ERROR_MESSAGE = (
    "`datasets' library is required to load ZTF M31 features dataset, install it with `pip install datasets` "
    "or reinstall the package with `pip install coniferest[datasets]`"
)


def ztf_m31(**kwargs):
    """
    Load ZTF DR3 M31 light curve feature dataset.

    The dataset is hosted on Hugging Face, `datasets' package is required
    to load it.

    Adopted from Malanchev et al. 2021:
    https://ui.adsabs.harvard.edu/abs/2021MNRAS.502.5147M/abstract
    https://zenodo.org/record/4318700

    Parameters
    ----------
    **kwargs
        Arbitrary keyword arguments passed to `datasets.load_dataset`.

    Returns
    -------
    data : 2-D numpy.ndarray of float32
        2-D array of light curve features.
    metadata : 1-D numpy.ndarray of uint64
        ZTF DR object IDs (OIDs) of the objects.
    """
    try:
        from datasets import load_dataset
    except ImportError:
        raise ImportError(IMPORT_ERROR_MESSAGE)

    dataset = load_dataset("snad-space/ztf-dr3-m31-features", **kwargs)
    df = dataset["train"].to_pandas()
    metadata = df["oid"].to_numpy(dtype=np.uint64)
    data = df.drop(columns=["oid"]).to_numpy(dtype=np.float32)
    return data, metadata



import math

import numpy as np

from .calc_trees import calc_apply, calc_feature_delta_sum, calc_paths_sum, selector_dtype  # noqa
from .utils import average_path_length

__all__ = ["ForestEvaluator"]


class ForestEvaluator:
    selector_dtype = selector_dtype

    def __init__(self, samples, selectors, node_offsets, leaf_offsets, *, num_threads, sampletrees_per_batch):
        """
        Base class for the forest evaluators. Does the trivial job:
        * runs calc_paths_sum written in Rust,
        * helps to combine several trees' representations into two arrays.

        Parameters
        ----------
        samples
            Number of samples in every tree.

        selectors
            Array with all the nodes of all the trees.

        node_offsets
            Offsets for tree indices for node-sized arrays. For example
            for two trees of node-length `len1` and `len2` it would be:
            [0, len1, len1 + len2]

        leaf_offsets
            Offsets for tree indices for leaf-sized arrays.

        num_threads : int or None
            Number of threads to use for calculations. If None then
        """
        self.samples = samples

        self.selectors = selectors
        self.node_offsets = node_offsets
        self.leaf_offsets = leaf_offsets

        if num_threads is None or num_threads < 0:
            # Ask Rust's rayon to use all available threads
            self.num_threads = 0
        else:
            self.num_threads = num_threads

        self.sampletrees_per_batch = sampletrees_per_batch

    @property
    def batch_size(self):
        return math.ceil(self.sampletrees_per_batch / self.n_trees)

    @classmethod
    def combine_selectors(cls, selectors_list):
        """
        Combine several node arrays into one array of nodes and one array of
        start node_offsets.

        Parameters
        ----------
        selectors_list
            List of node arrays to combine.

        Returns
        -------
        np.ndarray of selectors
            Node array with all the nodes from all the trees.
        np.ndarray of int
            Array of tree offsets for node-arrays.
        np.ndarray of int
            Array of tree offsets for leaf-arrays.
        """
        lens = [len(sels) for sels in selectors_list]
        full_len = sum(lens)

        selectors = np.empty((full_len,), dtype=cls.selector_dtype)

        node_offsets = np.zeros((len(selectors_list) + 1,), dtype=np.uintp)
        node_offsets[1:] = np.add.accumulate(lens)

        for i in range(len(selectors_list)):
            selectors[node_offsets[i] : node_offsets[i + 1]] = selectors_list[i]

        # Assign a unique sequential index to every leaf
        # The index is used for weighted scores
        leaf_mask = selectors["feature"] < 0

        # Each offset tells how many leafs are in all previous trees
        leaf_offsets = np.zeros_like(node_offsets)
        leaf_offsets[1:] = np.cumsum(leaf_mask)[node_offsets[1:] - 1]
        leaf_count = leaf_offsets[-1]

        selectors["left"][leaf_mask] = np.arange(0, leaf_count)

        return selectors, node_offsets, leaf_offsets

    @property
    def n_trees(self):
        return self.node_offsets.shape[0] - 1

    @property
    def n_leaves(self):
        return self.leaf_offsets[-1]

    def score_samples(self, x):
        """
        Perform the computations.

        Parameters
        ----------
        x
            Features to calculate scores of. Should be C-contiguous for performance.

        Returns
        -------
        Array of scores.
        """
        if not x.flags["C_CONTIGUOUS"]:
            x = np.ascontiguousarray(x)

        return -(
            2
            ** (
                -calc_paths_sum(
                    self.selectors,
                    self.node_offsets,
                    x,
                    num_threads=self.num_threads,
                    batch_size=self.batch_size,
                )
                / (self.average_path_length(self.samples) * self.n_trees)
            )
        )

    def _feature_delta_sum(self, x):
        if not x.flags["C_CONTIGUOUS"]:
            x = np.ascontiguousarray(x)

        return calc_feature_delta_sum(
            self.selectors,
            self.node_offsets,
            x,
            num_threads=self.num_threads,
            batch_size=self.batch_size,
        )

    def feature_signature(self, x):
        delta_sum, hit_count = self._feature_delta_sum(x)

        return delta_sum / hit_count / self.average_path_length(self.samples)

    def feature_importance(self, x):
        delta_sum, hit_count = self._feature_delta_sum(x)

        return np.sum(delta_sum, axis=0) / np.sum(hit_count, axis=0) / self.average_path_length(self.samples)

    def apply(self, x):
        if not x.flags["C_CONTIGUOUS"]:
            x = np.ascontiguousarray(x)

        return calc_apply(
            self.selectors,
            self.node_offsets,
            x,
            num_threads=self.num_threads,
            batch_size=self.batch_size,
        )

    @classmethod
    def average_path_length(cls, n_nodes):
        """
        Average path length is abstracted because in different cases we may want to
        use a bit different formulas to make the exact match with other software.

        By default we use our own implementation.
        """
        return average_path_length(n_nodes)



from onnxconverter_common.data_types import FloatTensorType
from onnxconverter_common.topology import Topology

from ..aadforest import AADForest
from ._container import ConiferestModelContainer


def _get_coniferest_operator_name(model):
    if isinstance(model, AADForest):
        return "AADForest"

    raise ValueError("No proper operator name found for '%s'" % type(model))


def _parse_coniferest(scope, model, inputs):
    this_operator = scope.declare_local_operator(_get_coniferest_operator_name(model), model)
    this_operator.inputs = inputs

    # FIXME: probably another variable is required for anomality label
    score_variable = scope.declare_local_variable("score", FloatTensorType())

    this_operator.outputs.append(score_variable)

    return this_operator.outputs


def parse_coniferest(
    model,
    initial_types=None,
    target_opset=None,
    custom_conversion_functions=None,
    custom_shape_calculators=None,
):
    raw_model_container = ConiferestModelContainer(model)
    topology = Topology(
        raw_model_container,
        default_batch_size="None",
        initial_types=initial_types,
        target_opset=target_opset,
        custom_conversion_functions=custom_conversion_functions,
        custom_shape_calculators=custom_shape_calculators,
    )
    scope = topology.declare_scope("__root__")

    inputs = []
    for var_name, initial_type in initial_types:
        inputs.append(scope.declare_local_variable(var_name, initial_type))

    for variable in inputs:
        raw_model_container.add_input(variable)

    outputs = _parse_coniferest(scope, model, inputs)

    for variable in outputs:
        raw_model_container.add_output(variable)

    return topology



from abc import ABC, abstractmethod
from warnings import warn

import numpy as np
from sklearn.ensemble._bagging import _generate_indices  # noqa
from sklearn.tree._criterion import MSE  # noqa
from sklearn.tree._splitter import RandomSplitter  # noqa
from sklearn.tree._tree import DepthFirstTreeBuilder, Tree  # noqa
from sklearn.utils.validation import check_random_state

from .evaluator import ForestEvaluator
from .utils import average_path_length

__all__ = ["Coniferest", "ConiferestEvaluator"]

# Instead of doing:
# from sklearn.utils._random import RAND_R_MAX
# we have:
RAND_R_MAX = 0x7FFFFFFF


# Cause RAND_R_MAX is restricted to C-code.


class Coniferest(ABC):
    """
    Base class for the forests in the package. It settles the basic
    low-level machinery with the sklearn's trees, used here.

    Parameters
    ----------
    trees : list or None, optional
        List with the trees in the forest. If None, then empty list is used.

    n_subsamples : int, optional
        Subsamples to use for the training.

    max_depth : int or None, optional
        Maximum depth of the trees in use. If None, then `log2(n_subsamples)`

    n_jobs : int, optional
        Number of threads to use for scoring. If -1, then number of CPUs is used.

    random_seed : int or None, optional
        Seed for the reproducibility. If None, then random seed is used.
    """

    def __init__(
        self, trees=None, n_subsamples=256, max_depth=None, n_jobs=-1, random_seed=None, sampletrees_per_batch=1 << 20
    ):
        self.trees = trees or []
        self.n_subsamples = n_subsamples
        self.max_depth = max_depth or int(np.log2(n_subsamples))

        self.n_jobs = n_jobs
        self.sampletrees_per_batch = sampletrees_per_batch

        # For the better future with reproducible parallel tree building.
        # self.seedseq = np.random.SeedSequence(random_state)
        # rng, = self.seedseq.spawn(1)
        # self.rng = np.random.default_rng(rng)

        self.rng = np.random.default_rng(random_seed)

        # The following are the setting for the tree building procedures.

        # May we use the same data points during subsampling? No.
        self.bootstrap_samples = False
        # How many samples the node should have at least to perform a split? Two.
        self.min_samples_split = 2
        # How many samples the leaf might have? One.
        self.min_samples_leaf = 1
        # Don't know what it is.
        self.min_weight_leaf = 0
        # Don't know. Deprecated. And deleted in newer version of sklearn.
        self.min_impurity_decrease = 0
        # How many outputs does each experiment (point) have? Zero can't be in sklearn.
        self.n_outputs = 1

    def build_trees(self, data, n_trees):
        """
        Just build `n_trees` trees from supplied `data`.

        Parameters
        ----------
        data
            Features.

        n_trees
            Number of trees to build

        Returns
        -------
        List of trees.
        """
        n_population, n_features = data.shape

        n_samples = self.n_subsamples
        if n_samples > n_population:
            msg1 = "population should be greater or equal than subsamples number"
            msg2 = f"got n_population < n_subsamples ({n_population} < {n_samples})"
            msg3 = f"assuming n_subsamples = {n_population}"
            warn(msg1 + ", " + msg2 + ", " + msg3)
            n_samples = n_population

        trees = []
        for tree_index in range(n_trees):
            random_state = check_random_state(self.rng.integers(RAND_R_MAX))
            indices = _generate_indices(
                random_state=random_state,
                bootstrap=self.bootstrap_samples,
                n_population=n_population,
                n_samples=n_samples,
            )

            subsamples = data[indices, :]
            tree = self.build_one_tree(subsamples)
            trees.append(tree)

        return trees

    def build_one_tree(self, data):
        """
        Build just one tree.

        Parameters
        ----------
        data
            Features to build that one tree of.

        Returns
        -------
        A tree.
        """
        # Hollow plug
        criterion = MSE(self.n_outputs, self.n_subsamples)

        # Splitter for splitting the nodes.
        splitter_state = check_random_state(self.rng.integers(RAND_R_MAX))
        splitter = RandomSplitter(
            criterion=criterion,
            max_features=1,
            min_samples_leaf=self.min_samples_leaf,
            min_weight_leaf=self.min_weight_leaf,
            random_state=splitter_state,
            monotonic_cst=None,
        )

        builder_args = {
            "splitter": splitter,
            "min_samples_split": self.min_samples_split,
            "min_samples_leaf": self.min_samples_leaf,
            "min_weight_leaf": self.min_weight_leaf,
            "max_depth": self.max_depth,
            "min_impurity_decrease": self.min_impurity_decrease,
        }

        # Initialize the builder
        builder = DepthFirstTreeBuilder(**builder_args)

        # Initialize the tree
        n_samples, n_features = data.shape
        tree = Tree(n_features, np.array([1] * self.n_outputs, dtype=np.int64), self.n_outputs)

        # Cause of sklearn bugs we cannot do this:
        # y = np.zeros((n_samples, self.n_outputs))
        # Instead we do:
        y = np.empty((n_samples, self.n_outputs))
        y_column = np.arange(n_samples)
        for oi in range(self.n_outputs):
            y[:, oi] = y_column
        # The counterpart is rnd.uniform from sklearn.ensemble.IsolationForest.fit.

        # And finally build that tree.
        builder.build(tree, data, y)

        return tree

    @staticmethod
    def _validate_known_data(known_data=None, known_labels=None):
        known_data = np.asarray(known_data) if known_data is not None else None
        known_labels = np.asarray(known_labels) if known_labels is not None else None

        if (known_data is None) != (known_labels is None):
            raise ValueError("known_data and known_labels must be provided together or both be None")

        if (known_data is not None) and len(known_data) != len(known_labels):
            raise ValueError(
                f"known_data and known_labels must have the same length: {len(known_data)} != {len(known_labels)}"
            )

        return known_data, known_labels

    @abstractmethod
    def fit(self, data, labels=None):
        """
        Fit to the applied data.
        """
        raise NotImplementedError()

    @abstractmethod
    def fit_known(self, data, known_data=None, known_labels=None):
        """
        Fit to the applied data with priors.
        """
        raise NotImplementedError()

    @abstractmethod
    def score_samples(self, samples):
        """
        Evaluate scores for samples.
        """
        raise NotImplementedError()

    @abstractmethod
    def feature_signature(self, x):
        raise NotImplementedError()

    @abstractmethod
    def feature_importance(self, x):
        raise NotImplementedError()


class ConiferestEvaluator(ForestEvaluator):
    """
    Fast evaluator of scores for Coniferests.

    Parameters
    ----------
    coniferest : Coniferest
        The forest for building the evaluator from.
    map_value : callable or None
        Optional function to map leaf values, mast accept 1-D array of values
        and return an array of the same shape.
    """

    def __init__(self, coniferest, map_value=None):
        selectors_list = [self.extract_selectors(t, map_value) for t in coniferest.trees]
        selectors, node_offsets, leaf_offsets = self.combine_selectors(selectors_list)

        super().__init__(
            samples=coniferest.n_subsamples,
            selectors=selectors,
            node_offsets=node_offsets,
            leaf_offsets=leaf_offsets,
            num_threads=coniferest.n_jobs,
            sampletrees_per_batch=coniferest.sampletrees_per_batch,
        )

    @classmethod
    def extract_selectors(cls, tree, map_value=None):
        """
        Extract node representations for the tree.

        Parameters
        ----------
        tree
            Tree to extract selectors from.
        map_value
            Optional function to map leaf values

        Returns
        -------
        Array with selectors.
        """
        nodes = tree.__getstate__()["nodes"]
        selectors = np.zeros_like(nodes, dtype=cls.selector_dtype)

        selectors["feature"] = nodes["feature"]
        selectors["feature"][selectors["feature"] < 0] = -1

        selectors["left"] = nodes["left_child"]
        selectors["right"] = nodes["right_child"]
        selectors["value"] = nodes["threshold"]

        n_node_samples = nodes["n_node_samples"]

        selectors["log_n_node_samples"] = np.log(n_node_samples)

        def correct_values(i, depth):
            if selectors[i]["feature"] < 0:
                value = depth + average_path_length(n_node_samples[i])
                selectors[i]["value"] = value if map_value is None else map_value(value)
            else:
                correct_values(selectors[i]["left"], depth + 1)
                correct_values(selectors[i]["right"], depth + 1)

        correct_values(0, 0)

        return selectors



import numpy as np

from ..evaluator import ForestEvaluator


class IsolationForestEvaluator(ForestEvaluator):
    def __init__(self, isoforest):
        """
        Create evaluator for sklearn's version of isolation forest.

        Parameters
        ----------
        isoforest
            Sklearn's isolation forest instance.
        """
        selectors_list = [self.extract_selectors(e) for e in isoforest.estimators_]
        selectors, node_offsets, leaf_offsets = self.combine_selectors(selectors_list)

        super(IsolationForestEvaluator, self).__init__(
            samples=isoforest.max_samples_,
            selectors=selectors,
            node_offsets=node_offsets,
            leaf_offsets=leaf_offsets,
            num_threads=isoforest.n_jobs,
            sampletrees_per_batch=1 << 20,
        )

    @classmethod
    def extract_selectors(cls, estimator):
        nodes = estimator.tree_.__getstate__()["nodes"]
        selectors = np.zeros_like(nodes, dtype=cls.selector_dtype)

        selectors["feature"] = nodes["feature"]
        selectors["feature"][selectors["feature"] < 0] = -1

        selectors["left"] = nodes["left_child"]
        selectors["right"] = nodes["right_child"]
        selectors["value"] = nodes["threshold"]

        n_node_samples = nodes["n_node_samples"]

        def correct_values(i, depth):
            if selectors[i]["feature"] < 0:
                selectors[i]["value"] = depth + cls.average_path_length(n_node_samples[i])
            else:
                correct_values(selectors[i]["left"], depth + 1)
                correct_values(selectors[i]["right"], depth + 1)

        correct_values(0, 0)

        return selectors

    @classmethod
    def average_path_length(cls, n):
        """
        Our average_path_length is a bit different from sklearn's one.
        So we reproduce the sklearn's realization here.
        """
        if n <= 1:
            return 0
        elif n == 2:
            return 1
        else:
            return 2.0 * (np.log(n - 1.0) + np.euler_gamma) - 2.0 * (n - 1.0) / n



from numbers import Real
from typing import Callable

import clarabel
import numpy as np
from scipy import sparse

from .calc_trees import calc_paths_sum  # noqa
from .coniferest import Coniferest, ConiferestEvaluator
from .label import Label

__all__ = ["AADForest"]


class AADEvaluator(ConiferestEvaluator):
    def __init__(self, aad):
        super(AADEvaluator, self).__init__(aad, map_value=aad.map_value)
        self.C_a = aad.C_a
        self.budget = aad.budget
        self.prior_influence = aad.prior_influence
        self.weights = np.full(shape=(self.n_leaves,), fill_value=np.reciprocal(np.sqrt(self.n_leaves)))

        leaf_mask = self.selectors["feature"] < 0
        self.leaf_values = self.selectors["value"][leaf_mask]

    def _q_tau(self, scores):
        if isinstance(self.budget, int):
            if self.budget >= len(scores):
                return np.max(scores)

            return np.partition(scores, self.budget)[self.budget]
        elif isinstance(self.budget, float):
            return np.quantile(scores, self.budget)

        raise ValueError("self.budget must be an int or float")

    def score_samples(self, x, weights=None):
        """
        Perform the computations.

        Parameters
        ----------
        x
            Features to calculate scores of. Should be C-contiguous for performance.
        weights
            Specific leaf weights to use instead of self.weights

        Returns
        -------
        Array of scores.
        """
        if not x.flags["C_CONTIGUOUS"]:
            x = np.ascontiguousarray(x)

        if weights is None:
            weights = self.weights

        return calc_paths_sum(
            self.selectors,
            self.node_offsets,
            x,
            weights,
            num_threads=self.num_threads,
            batch_size=self.batch_size,
        )

    def fit_known(self, data, known_data, known_labels, prior_weights=None):
        scores = self.score_samples(data)
        q_tau = self._q_tau(scores)
        known_leafs = self.apply(known_data)

        n_anomalies = np.count_nonzero(known_labels == Label.ANOMALY)
        n_nominals = np.count_nonzero(known_labels == Label.REGULAR)
        prior_influence = self.prior_influence(n_anomalies, n_nominals)

        if prior_weights is None:
            prior_weights = self.weights

        n_weights = self.weights.shape[0]
        n_knowns = n_anomalies + n_nominals
        n_variables = n_weights + n_knowns
        n_constraints = 2 * n_knowns

        # Problem matrix P
        data = np.full_like(self.weights, prior_influence)
        ind = np.arange(n_weights)
        P = sparse.csc_matrix((data, (ind, ind)), shape=(n_variables, n_variables))

        # Problem vector q
        q_known = np.zeros_like(known_labels, dtype=self.weights.dtype)
        if n_anomalies > 0:
            q_known[known_labels == Label.ANOMALY] = self.C_a * np.reciprocal(float(n_anomalies))
        if n_nominals > 0:
            q_known[known_labels == Label.REGULAR] = np.reciprocal(float(n_nominals))

        q = np.concatenate([-prior_influence * prior_weights, q_known])

        # Constraints matrix A
        ## Weight variables
        col_ind = known_leafs.flatten()
        row_ind = np.repeat(np.arange(n_knowns), self.n_trees)
        data = (-self.leaf_values[known_leafs] * known_labels.reshape(-1, 1)).flatten()

        ## Cost variables
        col_ind = np.concatenate(
            [
                col_ind,
                np.tile(n_weights + np.arange(n_knowns), 2),
            ]
        )
        row_ind = np.concatenate(
            [
                row_ind,
                np.arange(n_constraints),
            ]
        )
        data = np.concatenate(
            [
                data,
                np.full((n_constraints,), -1),
            ]
        )
        A = sparse.csc_matrix((data, (row_ind, col_ind)), shape=(n_constraints, n_variables))

        # Constraints vector b
        b = np.concatenate(
            [
                -q_tau * known_labels,
                np.zeros((n_knowns,)),
            ]
        )

        # Constraints type: all \ge 0
        cones = [
            clarabel.NonnegativeConeT(n_constraints),
        ]

        settings = clarabel.DefaultSettings()
        settings.verbose = False

        solver = clarabel.DefaultSolver(P, q, A, b, cones, settings)

        res = solver.solve()

        self.weights = np.asarray(res.x[:n_weights])
        weights_norm = np.linalg.norm(self.weights)
        self.weights /= weights_norm


class AADForest(Coniferest):
    r"""
    Active Anomaly Detection with Isolation Forest.

    See Das et al., 2017 https://arxiv.org/abs/1708.09441

    The method solves the optimization problem:

    .. math::

       \mathbf{w} = \arg\min_{\mathbf{w}} \left(
        \frac{C_a}{\left|\mathcal{A}\right|}
            \sum_{i \in \mathcal{A}} \mathrm{ReLU}\left(s(\mathbf{x_i} | \mathbf{w}) - q_{\tau}\right) +
        \frac{1}{\left|\mathcal{N}\right|}
            \sum_{i \in \mathcal{N}} \mathrm{ReLU}\left(q_{\tau} - s(\mathbf{x_i} | \mathbf{w})\right) +
        \frac{1}{2} \lVert \mathbf{w} - \mathbf{w_0}\rVert^2\right),

    where :math:`C_a` is `C_a`, :math:`\mathcal{A}` is a set of known anomalies,
    :math:`\mathcal{N}` is a set of known nominals, :math:`s(\mathbf{x_i} | \mathbf{w})`
    is the anomaly score of instance with features :math:`\mathbf{x_i}`
    given weights  :math:`\mathbf{w}`.

    This problem is reformulated as an equivalent quadratic programming problem:

    .. math::

       \begin{bmatrix}
       \mathbf{w}\\
       \mathbf{u}
       \end{bmatrix} = \arg\min_{\mathbf{w}, \mathbf{u}} \left(
        \frac{C_a}{\left|\mathcal{A}\right|} \sum_{i \in \mathcal{A}} u_i +
        \frac{1}{\left|\mathcal{N}\right|} \sum_{i \in \mathcal{N}} u_i +
        \frac{1}{2} \lVert \mathbf{w} - \mathbf{w_0} \rVert^2\right),

    with the following convex constraints:

    .. math::

       u_i &\ge 0 \quad & i \in \mathcal{A} \cup \mathcal{N},\\
       u_i - s(\mathbf{x_i} | \mathbf{w}) &\ge - q_{\tau}\quad & i \in \mathcal{A},\\
       u_i + s(\mathbf{x_i} | \mathbf{w}) &\ge q_{\tau}\quad & i \in \mathcal{N}.\\

    Parameters
    ----------
    n_trees : int, optional
        Number of trees in the isolation forest.

    n_subsamples : int, optional
        How many subsamples should be used to build every tree.

    max_depth : int or None, optional
        Maximum depth of every tree. If None, `log2(n_subsamples)` is used.

    budget : int or float, optional
        Budget of anomalies. If the type is floating point it is considered as
        fraction of full data. If the type is integer it is considered as the
        number of items. Default is 0.03.

    n_jobs : int or None, optional
        Number of threads to use for scoring. If None - all available CPUs are used.

    random_seed : int or None, optional
        Random seed to use for reproducibility. If None - random seed is used.

    prior_influence : float or callable, optional
        An regularization coefficient value in the loss functioin. Default is 1.0.
        Signature: '(anomaly_count, nominal_count) -> float'

    map_value : ["const", "exponential", "linear", "reciprocal"] or callable, optional
        An function applied to the leaf depth before weighting. Possible
        meaning variants are: 1, 1-exp(-x), x, -1/x.
    """

    def __init__(
        self,
        n_trees=100,
        n_subsamples=256,
        max_depth=None,
        budget=0.03,
        C_a=1.0,
        prior_influence=1.0,
        n_jobs=None,
        random_seed=None,
        sampletrees_per_batch=1 << 20,
        map_value=None,
    ):
        super().__init__(
            trees=[],
            n_subsamples=n_subsamples,
            max_depth=max_depth,
            n_jobs=n_jobs,
            random_seed=random_seed,
            sampletrees_per_batch=sampletrees_per_batch,
        )
        self.n_trees = n_trees

        if not isinstance(budget, (int, float)):
            raise ValueError("budget must be an int or float")

        self.budget = budget
        self.C_a = C_a

        if isinstance(prior_influence, Callable):
            self.prior_influence = prior_influence
        elif isinstance(prior_influence, Real):
            self.prior_influence = lambda anomaly_count, nominal_count: prior_influence
        else:
            raise ValueError("prior_influence is neither a callable nor a constant")

        MAP_VALUES = {
            "const": np.ones_like,
            "exponential": lambda x: -np.expm1(-x),
            "linear": lambda x: x,
            "reciprocal": lambda x: -np.reciprocal(x),
        }

        if map_value is None:
            map_value = "reciprocal"

        if isinstance(map_value, Callable):
            self.map_value = map_value
        elif map_value in MAP_VALUES:
            self.map_value = MAP_VALUES[map_value]
        else:
            raise ValueError(f"map_value is neither a callable nor one of {', '.join(MAP_VALUES.keys())}.")

        self.evaluator = None

    def _build_trees(self, data):
        if len(self.trees) == 0:
            self.trees = self.build_trees(data, self.n_trees)
            self.evaluator = AADEvaluator(self)

    def fit(self, data, labels=None):
        """
        Build the trees with the data `data`.

        Parameters
        ----------
        data
            Array with feature values of objects.

        labels
            Optional. Labels of objects. May be regular, anomalous or unknown.
            See `Label` data for details.

        Returns
        -------
        self
        """
        # If no labels were supplied, train with them.
        if labels is None:
            return self.fit_known(data)

        # Otherwise select known data, and train on it.
        labels = np.asarray(labels)
        index = labels != Label.UNKNOWN
        return self.fit_known(data, data[index, :], labels[index])

    def fit_known(self, data, known_data=None, known_labels=None):
        """
        The same `fit` but with a bit of different API. Known data and labels
        are separated from training data for time and space optimality. High
        chances are that `known_data` is much smaller that `data`. At that case
        it is not reasonable to hold the labels for whole `data`.

        Parameters
        ----------
        data
            Training data (array with feature values) to build trees with.

        known_data
            Feature values of known data.

        known_labels
            Labels of known data.

        Returns
        -------
        self
        """

        known_data, known_labels = self._validate_known_data(known_data, known_labels)

        self._build_trees(data)

        if (
            known_data is None
            or len(known_data) == 0
            or known_labels is None
            or len(known_labels) == 0
            or np.all(known_labels == Label.UNKNOWN)
        ):
            return self

        index = known_labels != Label.UNKNOWN
        known_data, known_labels = known_data[index], known_labels[index]

        self.evaluator.fit_known(data, known_data, known_labels)

        return self

    def score_samples(self, samples):
        """
        Computer scores for the supplied data.

        Parameters
        ----------
        samples
            Feature values to compute scores on.

        Returns
        -------
        Array with computed scores.
        """
        return self.evaluator.score_samples(samples)

    def feature_signature(self, x):
        raise NotImplementedError()

    def feature_importance(self, x):
        raise NotImplementedError()

    def apply(self, x):
        """
        Apply the forest to X, return leaf indices.

        Parameters
        ----------
        x : ndarray shape (n_samples, n_features)
            2-d array with features.

        Returns
        -------
        x_leafs : ndarray of shape (n_samples, n_estimators)
            For each datapoint x in X and for each tree in the forest,
            return the index of the leaf x ends up in.
        """
        return self.evaluator.apply(x)



from uuid import uuid4

import numpy as np
import onnx
from onnxconverter_common.data_types import FloatTensorType
from onnxconverter_common.topology import convert_topology

from ..coniferest import Coniferest
from ._parse import parse_coniferest
from .proto import get_maximum_opset_supported


def _guess_initial_types(X, model, initial_types):
    if initial_types is not None:
        return initial_types

    if isinstance(X, np.ndarray):
        shape = [None, X.shape[1]]
        initial_types = [("X", FloatTensorType(shape))]

    if len(model.trees):
        shape = [None, model.trees[0].n_features]
        initial_types = [("X", FloatTensorType(shape))]

    return initial_types


def convert(
    model,
    name=None,
    initial_types=None,
    doc_string="",
    target_opset=None,
    targeted_onnx=onnx.__version__,
    custom_conversion_functions=None,
    custom_shape_calculators=None,
):
    """
    Convert coniferest model to ONNX.

    Parameters
    ----------
    model : Coniferest
        Coniferest model to convert.

    name : string
        ONNX model name.

    Returns
    -------
    ONNX model

    Examples
    --------
    >>> from coniferest.aadforest import AADForest
    >>> from coniferest.datasets import single_outlier
    >>> from coniferest.onnx import convert
    >>> from onnxconverter_common.data_types import FloatTensorType
    >>>
    >>> data, _metadata = single_outlier()
    >>> data = data.astype(np.float32)
    >>> model = AADForest().fit(data)
    >>> o = convert(model, initial_types = [('X', FloatTensorType([None, data.shape[1]]))])
    """

    if not isinstance(model, Coniferest):
        raise ValueError("model is not Coniferest object")

    if name is None:
        name = str(uuid4().hex)

    if initial_types is None:
        raise ValueError("Initial types are required. See usage of convert(...) in coniferest.onnx.convert for details")

    target_opset = target_opset if target_opset else get_maximum_opset_supported()

    topology = parse_coniferest(
        model,
        initial_types,
        target_opset,
        custom_conversion_functions,
        custom_shape_calculators,
    )
    topology.compile()

    onnx_model = convert_topology(topology, name, doc_string, target_opset, targeted_onnx)
    return onnx_model


def to_onnx(
    model,
    X=None,
    name=None,
    initial_types=None,
    target_opset=None,
    targeted_onnx=onnx.__version__,
    custom_conversion_functions=None,
    custom_shape_calculators=None,
):
    if name is None:
        name = "ONNX(%s)" % model.__class__.__name__

    initial_types = _guess_initial_types(X, model, initial_types)

    return convert(
        model,
        name=name,
        initial_types=initial_types,
        target_opset=target_opset,
        targeted_onnx=targeted_onnx,
        custom_conversion_functions=custom_conversion_functions,
        custom_shape_calculators=custom_shape_calculators,
    )


def save_onnx_model(model, filename=None):
    """
    Serialize ONNX model and save it to a file.

    Parameters
    ----------
    model
        ONNX model to serialize.

    filename
        filename or file object to save the model.

    Returns
    -------
    Serialized ONNX model as bytes
    """

    content = model.SerializeToString()
    if filename is not None:
        if hasattr(filename, "write"):
            filename.write(content)
        else:
            with open(filename, "wb") as f:
                f.write(content)
    return content



from typing import Callable, Dict, Optional

import numpy as np

from coniferest.coniferest import Coniferest
from coniferest.pineforest import PineForest

from ..label import Label
from .callback import prompt_decision_callback


class Session:
    """
    Active anomaly detection session

    Parameters
    ----------
    data : array-like, shape (n_samples, n_features), dtype is number
        2-D array of data points
    metadata : array-like, shape (n_samples,), dtype is any
        1-D array of metadata for each data point
    decision_callback : callable, optional
        Function to be called when expert decision is required, it must return
        `Label` object with the decision and may terminate the session via
        `Session.terminate()`. Default is `prompt_decision_callback`
        Signature: '(metadata, data, session) -> Label', where metadata is
        metadata of the object to be labeled, data is data of the object to be
        labeled, session is this session instance.
    on_refit_callbacks : list of callable, or callable, or None, optional
        Functions to be called when model is refitted (before
        "decision_callback"), default is empty list. This function may call
        `Session.terminate()`.
        Signature: '(session) -> None', where session is this session instance.
    on_decision_callbacks : list of callable, or callable, or None, optional
        Functions to be called when expert decision is made (after
        "decision_callback"), default is empty list. This function may call
        `Session.terminate()`.
        Signature: '(metadata, data, session) -> None', where metadata is
        metadata of the object has just been labeled, data is data of this
        object, session is this session instance.
    known_labels : dict, optional
        Dictionary of known anomaly labels, keys are data/metadata indices,
        values are labels of type `Label`. Default is empty dictionary.
    model : Coniferest or None, optional
        Anomaly detection model to use, default is `PineForest()`.

    Attributes
    ----------
    current : int
        Index of the last anomaly candidate
    last_decision : Label or None
        Label of the last anomaly candidate or None if no decision was made
    scores : array-like, shape (n_samples,)
        Current anomaly scores for all data points
    terminated : bool
        True if session is terminated
    known_labels : dict[int, Label]
        Current dictionary of known anomaly labels
    known_anomalies : array-like
        Array of indices of known anomalies
    known_regulars : array-like
        Array of indices of known regular objects
    known_unknowns : array-like
        Array of indices of known objects marked with `Label::UNKNOWN`
    model : Coniferest
        Anomaly detection model used

    Examples
    --------

    >>> from coniferest.datasets import ztf_m31
    >>> from coniferest.session import Label, Session
    >>> data, metadata = ztf_m31()
    >>> s = Session(
    ...    data=data,
    ...    metadata=metadata,
    ...    decision_callback=lambda *_: Label.ANOMALY,
    ...    on_decision_callbacks=[lambda _metadata, _data, session: session.terminate()],
    ... )
    >>> _ = s.run()
    >>> assert len(s.known_labels) == len(s.known_anomalies) == 1
    """

    @staticmethod
    def _prepare_callbacks(input_argument):
        if input_argument is None:
            callbacks = []
        elif isinstance(input_argument, list):
            callbacks = input_argument
        else:
            callbacks = [
                input_argument,
            ]

        if not all([isinstance(cb, Callable) for cb in callbacks]):
            raise ValueError("At least one of the callbacks is not callable")
        return callbacks

    @staticmethod
    def _invoke_callbacks(callbacks, *args, **kwargs):
        for cb in callbacks:
            cb(*args, **kwargs)

    def __init__(
        self,
        data,
        metadata,
        decision_callback=prompt_decision_callback,
        *,
        on_refit_callbacks=None,
        on_decision_callbacks=None,
        known_labels: Dict[int, Label] = None,
        model: Coniferest = None,
    ):
        self._data = np.atleast_2d(data)
        self._metadata = np.atleast_1d(metadata)

        if not isinstance(decision_callback, Callable):
            raise ValueError("decision_callback is not a callable")
        self._decision_cb = decision_callback

        try:
            self._on_refit_cb = self._prepare_callbacks(on_refit_callbacks)
        except ValueError:
            raise ValueError("on_refit_callbacks contains not callable object")

        try:
            self._on_decision_cb = self._prepare_callbacks(on_decision_callbacks)
        except ValueError:
            raise ValueError("on_decision_callbacks contains not callable object")

        if known_labels is None:
            self._known_labels = {}
        else:
            self._known_labels = dict(known_labels)

        if model is None:
            model = PineForest()
        if not isinstance(model, Coniferest):
            raise ValueError("model is not a Coniferest object")
        self._model = model

        self._scores = None
        self._current = None
        self._terminated = False

    def run(self) -> "Session":
        """Evaluate interactive anomaly detection session"""

        if self._terminated:
            raise RuntimeError("Session is already terminated")

        self.model.fit(self._data)

        while not self._terminated:
            known_data = self._data[list(self._known_labels.keys())]
            known_labels = np.fromiter(self._known_labels.values(), dtype=int, count=len(self._known_labels))
            self.model.fit_known(self._data, known_data, known_labels)

            self._invoke_callbacks(self._on_refit_cb, self)

            self._scores = self.model.score_samples(self._data)

            k = len(self._known_labels) + 1
            argtopk = self.argtopk_scores(k)
            self._current = None
            for ind in argtopk:
                if ind not in self._known_labels:
                    self._current = ind
                    break

            if self._current is None:
                self.terminate()
                break

            decision = self._decision_cb(self._metadata[self._current], self._data[self._current], self)
            self._known_labels[self._current] = decision

            self._invoke_callbacks(
                self._on_decision_cb,
                self._metadata[self._current],
                self._data[self._current],
                self,
            )

        return self

    def terminate(self) -> None:
        self._terminated = True

    def argtopk_scores(self, k: int) -> np.ndarray:
        if k >= self._scores.shape[0]:
            return np.argsort(self._scores)
        argtopk_unsort = np.argpartition(self._scores, k)[:k]
        argtopk = argtopk_unsort[np.argsort(self._scores[argtopk_unsort])]
        return argtopk

    @property
    def current(self) -> int:
        return self._current

    @property
    def last_decision(self) -> Optional[Label]:
        return self._known_labels.get(self._current, None)

    @property
    def scores(self) -> np.ndarray:
        return self._scores

    @property
    def known_labels(self) -> Dict[int, Label]:
        return self._known_labels

    @property
    def known_anomalies(self) -> np.ndarray:
        return np.array([idx for idx, label in self._known_labels.items() if label == Label.ANOMALY])

    @property
    def known_regulars(self) -> np.ndarray:
        return np.array([idx for idx, label in self._known_labels.items() if label == Label.REGULAR])

    @property
    def known_unknowns(self) -> np.ndarray:
        return np.array([idx for idx, label in self._known_labels.items() if label == Label.UNKNOWN])

    @property
    def model(self) -> Coniferest:
        return self._model

    @property
    def terminated(self) -> bool:
        return self._terminated



import numpy as np

from coniferest.label import Label

from .plasticc_gp import plasticc_gp
from .ztf_m31 import ztf_m31

__all__ = ["ztf_m31", "plasticc_gp", "single_outlier", "non_anomalous_outliers"]


class Dataset:
    def __init__(self, data, labels):
        """
        Dataset is an o-by-f array, where o is objects and f is features.
        """
        self.data = data
        self.labels = labels

    def to_data_metadata(self):
        return np.asarray(self.data), np.asarray(list(map(Label, self.labels)))


class SingleOutlierDataset(Dataset):
    def __init__(self, inliers=10_000, rng=0):
        rng = np.random.default_rng(rng)
        data_inliers = rng.normal(loc=0, scale=1, size=(inliers, 2))
        data_outlier = np.array([[1e6, -1e6]])
        data = np.vstack([data_inliers, data_outlier])
        labels = np.append(np.zeros(inliers), np.ones(1))
        super().__init__(data, labels)


def single_outlier(inliers=10_000, rng=0):
    return SingleOutlierDataset(inliers, rng).to_data_metadata()


class MalanchevDataset(Dataset):
    def __init__(self, inliers=1 << 10, outliers=1 << 5, regions=None, rng=0):
        """
        A simple dataset for testing the anomaly detection algorithms. It
        constits of one portion of regular data of `inliers` capacity, and
        three portions of outlier data of `outliers` capacity each. Every
        outlier portion maybe selected either as regular or anomalous. Example:

        ```
                MalanchevDataset(inliers=100, outliers=10, regions=(R,R,A))
             ┌───────────────────────────────────────────────────────────────┐
         1.12┤  .           .                                        .       │
             │        .  .         .                        .  .    .  .     │
         0.88┤.   . .        .                                      .      . │
             │   .                                             .             │
             │                                                               │
         0.64┤                                                               │
             │                .     .                                        │
             │         ... ..  .... ... .....                                │
          0.4┤        ....  .. .. .. .    .                                  │
             │          . ...     ..   ... .                                 │
         0.17┤        .  .  ...  ..... .  ..                   *             │
             │         .    .... .  . .. . .                 *  **           │
             │         .   .      . . . ...                 *     *         *│
        -0.07┤                                                       *       │
             │                                                               │
        -0.31┤                                               *               │
             └┬──────────────┬───────────────┬───────────────┬──────────────┬┘
             -0.2           0.16            0.53            0.89          1.26

        ```

        Here we have a plot of 100 inliers, 20 regular outliers (all plotted as
        dots) and 10 anomalous outliers (plotted as stars).
        """
        self.inliers = inliers
        self.outliers = outliers
        if regions is None:
            regions = np.array([Label.R, Label.R, Label.A])
        self.regions = regions

        rng = np.random.default_rng(rng)
        self.rng = rng

        x = np.concatenate(
            [
                self._generate_inliers(inliers, rng),
                self._generate_outliers(outliers, rng, [1, 1]),
                self._generate_outliers(outliers, rng, [0, 1]),
                self._generate_outliers(outliers, rng, [1, 0]),
            ]
        )

        x_labels = np.concatenate(
            [
                np.ones(inliers),
                self.regions[0] * np.ones(outliers),
                self.regions[1] * np.ones(outliers),
                self.regions[2] * np.ones(outliers),
            ]
        )

        super(MalanchevDataset, self).__init__(data=x, labels=x_labels)

    @staticmethod
    def _generate_inliers(n, rng):
        return rng.uniform([0, 0], [0.5, 0.5], (n, 2))

    @staticmethod
    def _generate_outliers(n, rng, loc=None):
        loc = loc or [1, 1]
        return rng.normal(loc, 0.1, (n, 2))


def non_anomalous_outliers(inliers=1 << 10, outliers=1 << 5, regions=None, seed=0):
    return MalanchevDataset(inliers, outliers, regions, seed).to_data_metadata()


class DevNetDataset(Dataset):
    """Deviation Network paper datasets.

    This class constructor would download datasets from the Deviation Network GitHub:
    https://github.com/GuansongPang/deviation-network

    It requires pandas to be installed.

    Arguments
    ---------
    name : str
        Name of the dataset to download. See `.avialble_datasets`.

    Attributes
    ----------
    avialble_datasets : list[str]
        List of available datasets to download.
    """

    _dataset_filenames = {
        "donors": "KDD2014_donors_10feat_nomissing_normalised.csv",
        "census": "census-income-full-mixed-binarized.tar.xz",
        "fraud": "creditcardfraud_normalised.tar.xz",
        "celeba": "celeba_baldvsnonbald_normalised.csv",
        "backdoor": "UNSW_NB15_traintest_backdoor.tar.xz",
        "campaign": "bank-additional-full_normalised.csv",
        "thyroid": "annthyroid_21feat_normalised.csv",
    }
    _dataset_urls = {
        name: f"https://github.com/GuansongPang/deviation-network/raw/master/dataset/{filename}"
        for name, filename in _dataset_filenames.items()
    }
    avialble_datasets = list(_dataset_filenames.keys())

    def __init__(self, name: str):
        try:
            import pandas as pd
        except ImportError:
            raise ImportError(
                "Pandas is required to load DevNet datasets, install it with `pip install pandas` or "
                "reinstall the package with `pip install coniferest[datasets]`"
            )

        if name not in self.avialble_datasets:
            raise ValueError(f"Dataset {name} is not available. Available datasets are: {self.avialble_datasets}")

        df = pd.read_csv(self._dataset_urls[name])

        # Last column is for class, the rest are features
        data = df.iloc[:, :-1].to_numpy(dtype=float)

        # In the original data, the labels are 1 for anomalies and 0 for regular data
        # We need 1 for regular data and -1 for anomalies
        labels = 1 - 2 * df.iloc[:, -1].to_numpy(dtype=int)

        super().__init__(data, labels)


def dev_net_dataset(name: str):
    f"""Download and return metadata and data for the Deviation Network dataset.
   
    This class constructor would download datasets from the Deviation Network GitHub:
    https://github.com/GuansongPang/deviation-network

    It requires pandas to be installed.
    
    Avialable datasets are: {", ".join(DevNetDataset.avialble_datasets)}

    Arguments
    ---------
    name : str
        Name of the dataset to download. See `.avialble_datasets`. 
    
    Returns
    -------
    data : array-like, shape (n_samples, n_features)
        2-D array of data points
    labels : array-like, shape (n_samples,)
        1-D array of `Label` objects for each data point
    """
    return DevNetDataset(name).to_data_metadata()
